
RESERVED_PROPERTY_NAMES = ("key", "rev", "_properties", "_db")

# NAMESPACE_STR = "namespace"
NAMESPACE_STR = "namespace"
TYPE_STR = "type"

FIELD_REQUIRED = "required"
FIELD_UPDATE_ACL = "update_acl"
